<?php
/**
 * Simply Schedule Appointments Availability Functions.
 *
 * @since   0.0.3
 * @package Simply_Schedule_Appointments
 */
use League\Period\Period;
use Cake\Chronos\Date;

/**
 * Simply Schedule Appointments Availability Functions.
 *
 * @since 0.0.3
 */
class SSA_Availability_Functions {
	/**
	 * Parent plugin class.
	 *
	 * @since 0.0.3
	 *
	 * @var   Simply_Schedule_Appointments
	 */
	protected $plugin = null;

	/**
	 * Constructor.
	 *
	 * @since  0.0.3
	 *
	 * @param  Simply_Schedule_Appointments $plugin Main plugin object.
	 */
	public function __construct( $plugin ) {
		$this->plugin = $plugin;
		$this->hooks();
	}

	/**
	 * Initiate our hooks.
	 *
	 * @since  0.0.3
	 */
	public function hooks() {
		// add_action( 'init', array( $this, 'debug' ) );
		add_filter( 'ssa/buffer_period/start_date', array( $this, 'get_period_with_buffered_start_date' ), 10, 3 );
		add_filter( 'ssa/buffer_period/end_date', array( $this, 'get_period_with_buffered_end_date' ), 10, 3 );
	}

	public function debug() {
		// $this->get_default_available_periods( 1, '2017-10-01', '2017-10-07' );
		$bookable_appointments = $this->get_bookable_appointments( 1, array(
			'start_date_min' => gmdate( 'Y-m-d 12:15:00' ),
			'end_date_max' => gmdate( 'Y-m-31 18:15:00', time() ),
		) );
		exit();
	}

	public function combine_abutting_periods( $periods_array ) {
		if ( !is_array( $periods_array ) ) {
			return $periods_array;
		}

		$combined_periods = array();
		while( count( $periods_array ) ) {
			$next_period = array_shift( $periods_array );
			if ( count( $combined_periods ) === 0 ) {
				$combined_periods[] = $next_period;
				continue;
			}

			$last_period = current(array_slice($combined_periods, -1));
			if ( $last_period->abuts( $next_period ) ) {
				if ( $last_period->isAfter( $next_period ) ) {
					$tmp_last_period = $last_period;
					$last_period = $next_period;
					$next_period = $tmp_last_period;
				}

				array_pop( $combined_periods );
				$combined_periods[] = new Period( $last_period->getStartDate(), $next_period->getEndDate() );
				continue;
			}

			$combined_periods[] = $next_period;
			continue;
		}

		$combined_periods = array_filter($combined_periods, function($value) { return $value !== ''; });
		return $combined_periods;
	}

	public function get_all_available_periods( $appointment_type, $args=array() ) {
		if ( (int)$appointment_type == $appointment_type ) {
			$appointment_type = $this->plugin->appointment_type_model->get( $appointment_type );
		}
		if ( empty( $appointment_type['id'] ) ) {
			return new WP_Error();
		}

		$args = shortcode_atts( array(
			'start_date_min' => '',
			
			'start_date_max' => '',
			'end_date_min' => '',
			'end_date_max' => '',
		), $args );


		if ( $this->plugin->settings_installed->is_enabled( 'advanced_scheduling' ) ) {
			if ( ! empty( $appointment_type['availability_start_date'] ) && $appointment_type['availability_start_date'] !== '0000-00-00 00:00:00' && ( empty( $args['start_date_min'] ) || $args['start_date_min'] < $appointment_type['availability_start_date'] ) ) {
				$args['start_date_min'] = $appointment_type['availability_start_date'];
			}
		}


		if ( $this->plugin->settings_installed->is_enabled( 'advanced_scheduling' ) ) {
			if ( ! empty( $appointment_type['availability_end_date'] ) && $appointment_type['availability_end_date'] !== '0000-00-00 00:00:00' && ( empty( $args['start_date_max'] ) || $args['start_date_max'] > $appointment_type['availability_end_date'] ) ) {
				$args['start_date_max'] = $appointment_type['availability_end_date'];
				if ( $args['start_date_min'] > $args['start_date_max'] ) {
					return array();
				}
			}
		}

		if ( empty( $appointment_type['availability_type'] ) || $appointment_type['availability_type'] != 'custom ') {
			$available_periods = $this->get_default_available_periods( $appointment_type, $args );
		} else {
			// TODO: support custom available periods
			// $available_periods = $this->get_custom_available_periods( $appointment_type['id'], $args );
		}
		return $available_periods;
	}

	public function get_custom_available_periods( $appointment_type, $args=array() ) {
		if ( (int)$appointment_type == $appointment_type ) {
			$appointment_type = $this->plugin->appointment_type_model->get( $appointment_type );
		}
		if ( empty( $appointment_type['id'] ) ) {
			return new WP_Error();
		}

		$args = shortcode_atts( array(
			'start_date_min' => '',
			'start_date_max' => '',
			'end_date_min' => '',
			'end_date_max' => '',
		), $args );

		$custom_available_blocks = $this->plugin->availability_model->query( array_merge( array(
			'appointment_type_id' => $appointment_type['id'],
			'is_available' => 1,
			'orderby' => 'start_date',
			'order' => 'ASC',
		), $args ) );

		$custom_available_periods = array();
		foreach ($custom_available_blocks as $key => $custom_available_block) {
			$custom_available_periods[] = new Period( $custom_available_block['start_date'], $custom_available_block['end_date'] );
		}

		$custom_available_periods = $this->combine_abutting_periods( $custom_available_periods );
		return $custom_available_periods;
	}

	public function get_default_available_periods( $appointment_type, $args ) {
		if ( (int)$appointment_type == $appointment_type ) {
			$appointment_type = $this->plugin->appointment_type_model->get( $appointment_type );
		}
		if ( empty( $appointment_type['id'] ) ) {
			return new WP_Error();
		}

		$args = shortcode_atts( array(
			'start_date_min' => '',
			'start_date_max' => '',
			'end_date_min' => '',
			'end_date_max' => '',
		), $args );

		$availability_home_timezone = $this->plugin->utils->get_datetimezone( $appointment_type['id'] );

		$default_available_periods = array();

		$start_date = new Date( $args['start_date_min'] );
		$end_date = new Date( $args['start_date_max'] );
		$end_date = $end_date->modify( '+1 day' ); // help avoid timezone issues, our get_bookable_appointments() function will remove ones we don't need				
		$range_period = new Period( $start_date, $end_date );
		foreach ($range_period->getDatePeriod( '1 DAY' ) as $datetime) {
			$day_of_week = $datetime->format( 'l' );
			if ( empty( $appointment_type['availability'][$day_of_week][0]['time_start'] ) ) {
				continue;
			}

			foreach ($appointment_type['availability'][$day_of_week] as $time_block ) {
				$start_date = SSA_Utils::get_datetime_in_utc( $datetime->format( 'Y-m-d '.$time_block['time_start'] ), $availability_home_timezone );
				$end_date = SSA_Utils::get_datetime_in_utc( $datetime->format( 'Y-m-d '.$time_block['time_end'] ), $availability_home_timezone );
				
				$new_period = new Period(
					$start_date,
					$end_date
				);

				$default_available_periods[] = $new_period;
			}
		}

		$default_available_periods = $this->combine_abutting_periods( $default_available_periods );
		return $default_available_periods;
	}

	public function get_booked_periods( $appointment_type, $args ) {
		$args = shortcode_atts( array(
			'start_date_min' => '',
			'start_date_max' => '',
			'end_date_min' => '',
			'end_date_max' => '',
			'buffered' => false,
		), $args );
		if ( (int)$appointment_type == $appointment_type ) {
			$appointment_type = $this->plugin->appointment_type_model->get( $appointment_type );
		}
		if ( empty( $appointment_type['id'] ) ) {
			return new WP_Error();
		}

		if ( apply_filters( 'ssa/get_booked_periods/should_separate_availability_for_appointment_types', false ) ) {
			$args['appointment_type_id'] = $appointment_type['id'];
		}

		$booked_blocks = $this->plugin->appointment_model->query( array_merge( array(
			'status' => 'booked',
		), $args ) );

		$booked_periods = array();
		foreach ($booked_blocks as $key => $booked_block) {
			$booked_period = new Period( $booked_block['start_date'], $booked_block['end_date'] );
			if ( !empty( $args['buffered'] ) ) {
				$booked_period = apply_filters( 'ssa/buffer_period/start_date', $booked_period, $booked_period, $appointment_type );	
				$booked_period = apply_filters( 'ssa/buffer_period/end_date', $booked_period, $booked_period, $appointment_type );	
				if ( empty( $booked_period ) ) {
					continue;
				}
			}
			$booked_periods[] = $booked_period;
		}

		$booked_periods = apply_filters( 'ssa/get_booked_periods/booked_periods', $booked_periods, $appointment_type );
		return $booked_periods;
	}

	public function get_blocked_periods( $appointment_type, $args ) {
		if ( (int)$appointment_type == $appointment_type ) {
			$appointment_type = $this->plugin->appointment_type_model->get( $appointment_type );
		}
		if ( empty( $appointment_type['id'] ) ) {
			return new WP_Error();
		}

		$args = shortcode_atts( array(
			'start_date_min' => '',
			'start_date_max' => '',
			'end_date_min' => '',
			'end_date_max' => '',
			'buffered' => false,
		), $args );

		$blocked_periods = array();

		// $blocked_blocks = $this->plugin->availability_model->query( array(
		// 	'appointment_type_id' => $appointment_type['id'],
		// 	'is_available' => 0,
		// 	'number' => -1,
		// 	// TODO: optimize query...
		// 	// calendar_id IN (specific,ids,from,appointment_type,config)
		// 	// 'start_date_min' => ...
		// 	// 'start_date_max' => ...
		// 	// 'end_date_min' => ...
		// 	// 'end_date_max' => ...
		// ) );

		// foreach ($blocked_blocks as $key => $blocked_block) {
		// 	$blocked_period = new Period( $blocked_block['start_date'], $blocked_block['end_date'] );
		// 	if ( !empty( $args['buffered'] ) ) {
		// 		$blocked_period = apply_filters( 'ssa/buffer_period/start_date', $blocked_period, $blocked_period, $appointment_type );
		// 		$blocked_period = apply_filters( 'ssa/buffer_period/end_date', $blocked_period, $blocked_period, $appointment_type );
		// 		if ( empty( $blocked_period ) ) {
		// 			continue;
		// 		}
		// 	}
		// 	$blocked_periods[] = $blocked_period;
		// }

		$blocked_periods = apply_filters( 'ssa/get_blocked_periods/blocked_periods', $blocked_periods, $appointment_type, $args );
		return $blocked_periods;
	}

	public function get_bookable_appointments( $appointment_type, $args=array() ) {
		if ( (int)$appointment_type == $appointment_type ) {
			$appointment_type = $this->plugin->appointment_type_model->get( $appointment_type );
		}
		if ( empty( $appointment_type['id'] ) ) {
			return new WP_Error();
		}

		$appointment_type = apply_filters( 'ssa/get_bookable_appointments/appointment_type', $appointment_type );

		$args = shortcode_atts( array(
			'start_date_min' => '',
			'start_date_max' => '',
			'end_date_min' => '',
			'end_date_max' => '',
		), $args );

		if ( !empty( $args['start_date_min'] ) ) {
			$start_date_min_datetime = new DateTimeImmutable( $args['start_date_min'] );
			if ( !empty( $appointment_type['min_booking_notice'] ) ) {
				$start_date_with_min_booking_notice_datetime = ssa_datetime()->add( new DateInterval( 'PT'.$appointment_type['min_booking_notice'].'M' ) );
				if ( $start_date_with_min_booking_notice_datetime > $start_date_min_datetime ) {
					$start_date_min_datetime = $start_date_with_min_booking_notice_datetime;
				}
			}
		}

		if ( !empty( $args['start_date_max'] ) ) {
			$start_date_max_datetime = new DateTimeImmutable( $args['start_date_max'] );
		}
		if ( !empty( $appointment_type['max_booking_notice'] ) ) {
			$start_date_with_max_booking_notice_datetime = ssa_datetime()->add( new DateInterval( 'PT'.$appointment_type['max_booking_notice'].'M' ) );
			if ( empty( $start_date_max_datetime ) || $start_date_with_max_booking_notice_datetime < $start_date_max_datetime ) {
				$start_date_max_datetime = $start_date_with_max_booking_notice_datetime;
			}
		}

		if ( empty( $appointment_type['availability_increment'] ) ) {
			$appointment_type['availability_increment'] = 15;
		}
		if ( !empty( $args['end_date_min'] ) ) {
			$end_date_min_datetime = new DateTimeImmutable( $args['end_date_min'] );
		}

		if ( !empty( $args['end_date_max'] ) ) {
			$end_date_max_datetime = new DateTimeImmutable( $args['end_date_max'] );
		}
		if ( !empty( $appointment_type['max_booking_notice'] ) ) {
			$end_date_with_max_booking_notice_datetime = ssa_datetime()->add( new DateInterval( 'PT'.$appointment_type['max_booking_notice'].'M' ) )->add( new DateInterval( 'PT'.$appointment_type['duration'].'M' ) );
			if ( empty( $end_date_max_datetime ) || $end_date_with_max_booking_notice_datetime < $end_date_max_datetime ) {
				$end_date_max_datetime = $end_date_with_max_booking_notice_datetime;
			}
		}


		$available_periods = $this->get_all_available_periods( $appointment_type['id'], $args );
		$booked_periods = $this->get_booked_periods( $appointment_type['id'], $args );
		$blocked_periods = $this->get_blocked_periods( $appointment_type['id'], $args );
		
		$bookable_periods = array();
		foreach ($available_periods as $key => $available_period) {
			foreach ($available_period->getDatePeriod($appointment_type['availability_increment'].' MIN') as $start_datetime) {
				/* Look at every bookable start time and determine if this slot is available */

				/* Does this potential appointment start within min/max datetime? */
				if ( !empty( $start_date_min_datetime ) && $start_datetime < $start_date_min_datetime ) {
					continue;
				}
				if ( !empty( $start_date_max_datetime ) && $start_datetime > $start_date_max_datetime ) {
					continue;
				}
				
				$bookable_period = Period::createFromDuration($start_datetime, new DateInterval('PT'.$appointment_type['duration'].'M'));

				/* Does this potential appointment end within min/max datetime? */
				if ( !empty( $end_date_min_datetime ) && $bookable_period->getEndDate() < $end_date_min_datetime ) {
					continue;
				}
				if ( !empty( $end_date_max_datetime ) && $bookable_period->getEndDate() > $end_date_max_datetime ) {
					continue;
				}

				/* Does this potential appointment fit within the available block? */
				if ( !$available_period->contains( $bookable_period ) ) {
					continue;
				}

				/* Does this potential appointment conflict with a blocked time? */
				$buffered_bookable_period = apply_filters( 'ssa/buffer_period/start_date', $bookable_period, $bookable_period, $appointment_type );
				$buffered_bookable_period = apply_filters( 'ssa/buffer_period/end_date', $buffered_bookable_period, $bookable_period, $appointment_type );

				foreach ($blocked_periods as $key => $blocked_period) {
					$buffered_blocked_period = apply_filters( 'ssa/buffer_period/blocked_period', $blocked_period, $blocked_period, $appointment_type );
					if ( $buffered_bookable_period->overlaps( $buffered_blocked_period ) ) {
						/* They overlap, but let's make sure that our double-buffering didn't exclude a valid bookable period (collapse margins) */
						if ( $bookable_period->overlaps( $blocked_period ) ) {
							// even unbuffered they overlap
							continue 2;
						}
						
						$gap = $bookable_period->gap( $blocked_period );
						$minimum_gap_in_seconds = 60 * max( absint( $appointment_type['buffer_before'] ), absint( $appointment_type['buffer_after'] ) );
						if ( $gap->getTimestampInterval() < $minimum_gap_in_seconds ) {
							continue 2;
						}
					}
				}

				/* Does this potential appointment conflict with a previously-booked time? */
				foreach ($booked_periods as $key => $booked_period) {
					$buffered_booked_period = apply_filters( 'ssa/buffer_period/start_date', $booked_period, $booked_period, $appointment_type );
					$buffered_booked_period = apply_filters( 'ssa/buffer_period/end_date', $buffered_booked_period, $booked_period, $appointment_type );
					if ( $buffered_bookable_period->overlaps( $buffered_booked_period ) ) {
						/* They overlap, but let's make sure that our double-buffering didn't exclude a valid bookable period (collapse margins) */
						if ( $bookable_period->overlaps( $booked_period ) ) {
							// even unbuffered they overlap
							continue 2;
						}
						
						$gap = $bookable_period->gap( $booked_period );
						$minimum_gap_in_seconds = 60 * max( absint( $appointment_type['buffer_before'] ), absint( $appointment_type['buffer_after'] ) );
						if ( $gap->getTimestampInterval() < $minimum_gap_in_seconds ) {
							continue 2;
						}
					}
				}

				$bookable_periods[] = $bookable_period;
			}
		}

		$bookable_appointments = array();
		foreach ($bookable_periods as $key => $bookable_period) {
			$bookable_appointments[] = array(
				'period' => $bookable_period,
			);
		}

		return $bookable_appointments;
	}

	public function get_period_with_buffered_start_date( $period, $original_period, $appointment_type ) {
		if ( !empty( $appointment_type['buffer_before'] ) ) {
			$buffer_before = '-' . absint( $appointment_type['buffer_before'] ) . ' MIN';
			$period = $period->moveStartDate( $buffer_before );
		}

		return $period;
	}

	public function get_period_with_buffered_end_date( $period, $original_period, $appointment_type ) {
		if ( !empty( $appointment_type['buffer_after'] ) ) {
			$buffer_after = '+' . absint( $appointment_type['buffer_after'] ) . ' MIN';
			$period = $period->moveEndDate( $buffer_after );
		}

		return $period;
	}

}
