<?php
/**
 * Simply Schedule Appointments Capabilities.
 *
 * @since   0.0.3
 * @package Simply_Schedule_Appointments
 */

/**
 * Simply Schedule Appointments Capabilities.
 *
 * @since 0.0.3
 */
class SSA_Capabilities {
	/**
	 * Parent plugin class.
	 *
	 * @since 0.0.3
	 *
	 * @var   Simply_Schedule_Appointments
	 */
	protected $plugin = null;

	/**
	 * Constructor.
	 *
	 * @since  0.0.3
	 *
	 * @param  Simply_Schedule_Appointments $plugin Main plugin object.
	 */
	public function __construct( $plugin ) {
		$this->plugin = $plugin;
		$this->hooks();
	}

	/**
	 * Initiate our hooks.
	 *
	 * @since  0.0.3
	 */
	public function hooks() {
		if ( self::has_members_plugin() ) {
			add_filter( 'members_get_capabilities', array( $this, 'filter_members_get_capabilities' ) );
		}

		add_filter( 'user_has_cap', array( 'SSA_Capabilities', 'user_has_cap' ), 10, 3 );
	}

	public static function has_members_plugin() {
		return function_exists( 'members_get_capabilities' );
	}

	// called by Member plugin filter. Provides the plugin with list of SSA capabilities
	public static function filter_members_get_capabilities( $caps ) {
		return array_merge( $caps, self::get_all_caps() );
	}

	public static function get_all_caps() {
		return array(
			'ssa_manage_settings' => __( 'SSA: Manage Settings', 'simply-schedule-appointments' ),
			'ssa_book_appointments' => __( 'SSA: Book Appointments', 'simply-schedule-appointments' ),
			'ssa_cancel_appointments' => __( 'SSA: Cancel Appointments', 'simply-schedule-appointments' ),
			'ssa_cancel_others_appointments' => __( 'SSA: Cancel Others\' Appointment', 'simply-schedule-appointments' ),
		);
	}

	public static function user_has_cap( $all_caps, $cap, $args ) {
		$capability = SSA_Utils::array_key( $cap, 0 );
		if ( strpos( $capability, 'ssa_' ) !== 0 ) {
			return $all_caps;
		}

		$ssa_caps = self::get_all_caps();
		if ( ! in_array( $capability, array_keys( $ssa_caps ) ) && $capability !== 'ssa_full_access' ) {
			return $all_caps;
		}

		if ( ! self::has_members_plugin() ) {
			//give full access to administrators if the members plugin is not installed
			if ( current_user_can( 'administrator' ) || is_super_admin() ) {
				$all_caps['ssa_full_access'] = true;
				$all_caps[$capability] = true;
			}
		} else if ( current_user_can( 'administrator' ) || is_super_admin() ) {
			//checking if user has any SSA permission.
			$has_ssa_cap = false;
			foreach ( $ssa_caps as $ssa_cap ) {
				if ( SSA_Utils::array_key( $all_caps, $ssa_cap ) ) {
					$has_ssa_cap = true;
				}
			}

			if ( ! $has_ssa_cap ) {
				//give full access to administrators if none of the SSA permissions are active by the Members plugin
				$all_caps['ssa_full_access'] = true;
				$all_caps[$capability] = true;
			}
		}

		return $all_caps;
	}

	public static function current_user_can_any( $caps ) {
		return self::user_can_any( get_current_user_id(), $caps );
	}

	public static function current_user_can_which( $caps ) {
		return self::user_can_which( get_current_user_id(), $caps );
	}

	public static function user_can_any( $user, $caps ) {

		if ( ! is_array( $caps ) ) {
			$has_cap = user_can( $user, $caps ) || user_can( $user, 'ssa_full_access' );

			return $has_cap;
		}

		foreach ( $caps as $cap ) {
			if ( user_can( $user, $cap ) ) {
				return true;
			}
		}

		$has_full_access = user_can( $user, 'ssa_full_access' );

		return $has_full_access;
	}

	public static function user_can_which( $user, $caps ) {

		foreach ( $caps as $cap ) {
			if ( user_can( $user, $cap ) ) {
				return $cap;
			}
		}

		return '';
	}

}
